# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Commands editor.
"""

#debug
#import pdb

#application modules
import cacaeditor
from cacaeditor.helpers import CommandObject
from cacaeditor.helpers import DrawImport, DrawText
from cacaeditor.widgets import MsgDialog, MsgError

#
# General commands
#

class CmdShowHelp(CommandObject):
    """ Command to show help message.

        Usage: help <command>
    """
    cmd = "help"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #reset cursor
        self.app.cursor.display(False)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #check for argument
        if args:
            cmd_list = [item.cmd for item in self.app.commands]
            if not args[0] in cmd_list:
                #show error dialog
                self.app.errdialog = MsgError("Unknown command '%s'." % args[0])
                self.app.mode.set(self.app.mode._last, 'ERRDIALOG')

                #interrupt
                return None
            else:
                #show help command dialog
                idx = cmd_list.index(args[0])
                self.app.dialog = MsgDialog(self.app.commands[idx].help(),
                                            width=70,
                                            wrap=False,
                                            title="HELP for `%s'" % args[0])
        else:
            #show help dialog
            self.app.dialog = MsgDialog(cacaeditor.HELP_MSG,
                                        wrap=70,
                                        title="HELP")

        #change application mode
        self.app.mode.set(self.app.mode._last, 'DIALOG')

class CmdShowCommands(CommandObject):
    """ Command to show all valid commands.

        Usage: list
    """
    cmd = "list"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #show commands list dialog
        cmd_list = [item.cmd for item in self.app.commands]
        cmd_list.sort()
        self.app.dialog = MsgDialog("%s." % ", ".join(cmd_list),
                                    wrap=70,
                                    title="AVAILABLE COMMANDS")

        #change application mode
        self.app.mode.revert()
        self.app.mode.set('DIALOG')
        self.app.cursor.display(False)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

class CmdShowAbout(CommandObject):
    """ Command to show about dialog.

        Usage: about
    """
    cmd = "about"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        self.app.cursor.display(False)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #call app action about
        self.app.action_about()

class CmdShowFormat(CommandObject):
    """ Command to show list format dialog.

        Usage: format
    """
    cmd = "format"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        self.app.cursor.display(False)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #call app action list format
        self.app.action_list_format()

#
# Document commands
#

class CmdNewDocument(CommandObject):
    """ Create new document.

        Usage: new
    """
    cmd = "new"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #create new document
        if self.app.mode() == 'READY':
            self.app.action_new_document()

class CmdOpenDocument(CommandObject):
    """ Open a document.

        Usage: open <filename> [format]

        filename = a valid path for file to open
        format   = a valid libcaca format (caca, text, ansi, utf8)
    """
    cmd = "open"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #create new document
        if self.app.mode() == 'READY':
            try:
                if args:
                    self.app.action_new_document()
                if args and len(args) == 1:
                    self.app.body.load(args[0])
                elif args and len(args) > 1:
                    self.app.body.load(args[0], args[1])
                self.app.cursor.display(True)
            except cacaeditor.FileError, err:
                #display message error
                self.app.cursor.display(False)
                self.app.errdialog = MsgError(str(err))
                self.app.mode.set('READY', 'ERRDIALOG')

class CmdCloseDocument(CommandObject):
    """ Close current document.

        Usage: close
    """
    cmd = "close"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #create new document
        if self.app.mode() == 'OPENED':
            self.app.action_close_document()

class CmdSaveDocument(CommandObject):
    """ Save a document.

        Usage: save [filename]

        filename = valid path to save document
    """
    cmd = "save"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #save document
        if self.app.mode() == 'OPENED':
            try:
                if not args:
                    self.app.body.save()
                else:
                    self.app.body.save(args[0])
            except cacaeditor.FileError, err:
                #display message error
                self.app.cursor.display(False)
                self.app.errdialog = MsgError(str(err))
                self.app.mode.set('ERRDIALOG')

class CmdImportDocument(CommandObject):
    """ Import a document.

        Usage: import <x> <y> <filename> [format]

        x = X coordinate for data
        y = Y coordinate for data
        filename = a valid path for file to import
        format   = a valid libcaca format (caca, text, ansi, utf8)
    """
    cmd = "import"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #change application mode
        self.app.mode.revert()
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        if self.app.mode() == 'OPENED':
            #parse arguments
            try:
                x = int(args[0])
                y = int(args[1])
                filename = args[2]
            except (IndexError, ValueError):
                #display error message
                self.app.cursor.display(False)
                self.app.errdialog = MsgError("Invalid argument")
                self.app.mode.set('ERRDIALOG')
                return None

            #optional argument
            if len(args) > 3:
                fmt = args[3]
            else:
                fmt = None

            #draw import data
            if self.is_on_canvas(x, y):
                try:
                    obj = DrawImport(x, y)
                    obj.load(filename, fmt)
                    self.app.body.blit(obj.x, obj.y, obj.cv)
                except cacaeditor.FileError, err:
                    #display error message
                    self.app.cursor.display(False)
                    self.app.errdialog = MsgError(str(err))
                    self.app.mode.set('ERRDIALOG')
                    return None

class CmdExportDocument(CommandObject):
    """ Export a document

        Usage: export <filename> <format>

        filename = a valid path for file to export
        format   = a valid libcaca format (caca, text, ansi, utf8)
    """
    cmd = "export"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        pass

#
# Drawing commands
#

class CmdChangeColor(CommandObject):
    """ Change foreground or background color or reset colors.

        Usage: color <reset|colorname> <fg|bg>

        reset     = reset to default colors
        colorname = a valid color
        fg | bg   = foreground or background

        (black, blue, brown, cyan, drakgray, green, lightblue,
        lightcyan, lightgreen, lightgray, lightmagenta, lightred,
        magenta, red, white, yellow)
    """
    cmd = "color"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #reset application mode
        self.app.mode.revert()
        self.app.cursor.display(True)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #change colors
        try:
            if args and args[0] == 'reset':
                self.app.colors.reset()
            elif args and args[0] in cacaeditor.COLOR_NAMES:
                if args[1] == 'fg':
                    self.app.colors.change(fg=cacaeditor.COLOR_NAMES[args[0]])
                elif args[1] == 'bg':
                    self.app.colors.change(bg=cacaeditor.COLOR_NAMES[args[0]])
        except IndexError:
            pass

class CmdDrawText(CommandObject):
    """ Insert text at coords.

        Usage: text <x> <y> <string>
    """
    cmd = "text"

    def launch(self, *args, **kw):
        """ Command launcher.
        """
        #reset application mode
        self.app.mode.revert()
        self.app.cursor.display(True)
        x, y = self.app.cursor.old_position
        self.app.cursor.set_position(x, y)

        #parse arguments
        try:
            x = int(args[0])
            y = int(args[1])
            text = args[2:]
        except IndexError:
            return None

        #draw text
        if self.is_on_canvas(x, y):
            obj = DrawText(x, y)
            obj.answer = [c for c in " ".join(text)]
            cv = obj.make_cv()
            self.app.body.blit(x, y, cv)

