# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Application model.
"""

#debug
#import pdb

#standard module
import io

#third party modules
import caca
from caca.canvas import Canvas, CanvasError, NullCanvas
from caca.display import Display, DisplayError, Event

#application modules
import cacaeditor
import cacaeditor.commands
from cacaeditor.helpers import ColorObject, CursorObject
from cacaeditor.helpers import ModeObject
from cacaeditor.helpers import debug
from cacaeditor.widgets import Buffer, MsgDialog, MsgError

class ModeApp(object):
    """ Model class to manage cacaeditor. The main application
        object inherit from this class.

        * build main interface and methods to manipulate it
        * dispatch all events from application (keyboard, mouse, resize...)
        * Enter main application loop
    """
    #set to True to end application
    quit_app = False

    def __init__(self, **kw):
        """ Event application constructor.

            Attributes:
                name        => String to represent application name
                cursor      => CursorObject to store cursor position
                mode        => ModeObject to store application mode
                colors      => ColorObject to store application colors
                commands    => List to store command objects
                menu        => MenuBar widget
                status      => StatusBar widget
                prompt      => List to store prompt value
                body        => Buffer widget
                dialog      => Dialog widget
                userdata    => List to store user data input
        """
        #set interface attributes
        self.name = "CacaEditor v%s" % cacaeditor.__version__
        self.mode = ModeObject()
        self.menu = None
        self.status = None
        self.prompt = ['>', ' ']
        self.body = None
        self.dialog = None
        self.errdialog = None
        self.userdata = []
        self.commands = []

        #drawing attributes
        self.drawobj = None

        #lock minimal window size
        column, row = 80, 30
        #parse options
        if kw.has_key('column'):
            if kw['column'] > column:
                column = kw['column']
        if kw.has_key('row'):
            if kw['row'] > row:
                row = kw['row']

        #enable debug mode
        if kw.has_key('debug'):
            cacaeditor.DEBUG = kw['debug']

        #update modules list to debug
        if kw.has_key('debug_modules'):
            if kw['debug_modules']:
                cacaeditor.DEBUG_MODULES = kw['debug_modules']

        #init libcaca display
        self.mode.set('INIT')

        try:
            self.cv = Canvas(column, row)
            self.dp = Display(self.cv)
        except (DisplayError, CanvasError), e:
            raise cacaeditor.AppError(e)

        #debug canvas
        debug(__name__, "Canvas: %dx%d, Driver: %s" % (self.cv.get_width(),
                                                       self.cv.get_height(),
                                                       self.dp.get_driver()))

        #set cursor with reference canvas & display
        self.cursor = CursorObject(self.cv, self.dp, 0, 1)

        #set color with reference canvas
        self.colors = ColorObject(self.cv,
                                  caca.COLOR_DEFAULT,
                                  caca.COLOR_TRANSPARENT)

        #overwrite caca.canvas.Canvas.draw_cp437_box()
        #hack to display thin border with slang driver
        if self.dp.get_driver() == 'slang':
            caca.canvas.Canvas.draw_cp437_box = caca.canvas.Canvas.draw_thin_box

        #build command list
        for mod in dir(cacaeditor.commands):
            if mod.startswith('Cmd'):
                obj = eval("cacaeditor.commands.%s" % mod)
                if obj.cmd in [item.cmd for item in self.commands]:
                    debug(__name__, "Skip command `%s': `%s' already registred"\
                          % (mod, obj.cmd))
                else:
                    self.commands.append(obj(self))
                    debug(__name__, "Register command `%s' as `%s'" \
                          % (mod, obj.cmd))

        #build interface
        self.build()

    def build(self):
        """ Build interface. Overwrite this method to render the
            interface, see: gui.py.
        """
        pass

    def cancel_action(self):
        """ Cancel user action.
        """
        #reset cursor
        x, y = self.cursor.old_position
        self.cursor.set_position(x, y)

        #reset application mode
        self.mode.revert()

    def center(self, widget):
        """ Return centered blit position for widget.
        """
        x = (self.cv.get_width() - widget.get_width()) / 2
        y = (self.cv.get_height() - widget.get_height()) / 2

        return x, y

    def destroy(self):
        """ Exit event loop and destroy application
        """
        self.quit_app = True

    def main_loop(self):
        """ Enter the main event loop.
        """
        ev = Event()
        while self.quit_app is False:
            #wait for an event
            self.dp.get_event(caca.EVENT_ANY, ev, -1)

            if ev.get_type() != caca.EVENT_KEY_RELEASE:
                #keyboard events
                if ev.get_type() == caca.EVENT_KEY_PRESS:
                    debug(__name__, "EVENT_KEY_PRESS triggered")
                    self.dispatch_keyboard_event(ev)

                ## mouse events
                elif ev.get_type() == caca.EVENT_MOUSE_MOTION:
                    debug(__name__, "EVENT_MOUSE_MOTION triggered")
                    self.dispatch_mouse_event(ev)
                elif ev.get_type() == caca.EVENT_MOUSE_PRESS:
                    debug(__name__, "EVENT_MOUSE_PRESS triggered")
                    self.dispatch_mouse_event(ev)
                elif ev.get_type() == caca.EVENT_MOUSE_RELEASE:
                    debug(__name__, "EVENT_MOUSE_RELEASE triggered")
                    self.dispatch_mouse_event(ev)

                ## resize events
                elif ev.get_type() == caca.EVENT_RESIZE:
                    debug(__name__, "EVENT_RESIZE triggered")
                    self.dispatch_resize_event(ev)

                ## quit events
                elif ev.get_type() == caca.EVENT_QUIT:
                    debug(__name__, "EVENT_QUIT triggered")
                    self.destroy()

                #refresh application
                debug(__name__, "Refresh display")
                self.refresh()

    def move_cursor(self, key):
        """ Move the cursor.

            :key:   the arrow key
        """
        #grab document screen size
        width = self.cv.get_width() - 1
        height = self.cv.get_height() - 2

        #grab document size
        body_width = self.body.cv.get_width()
        body_height = self.body.cv.get_height()

        #grab cursor position
        x, y = self.cursor.where()

        #compute new cursor position
        if key == caca.KEY_UP:
            y -= 1
            if y < 1:
                y = 1
                self.body.posy -= 1
                if self.body.posy < 0: self.body.posy = 0
        elif key == caca.KEY_DOWN:
            y += 1
            if y > height:
                y = height
                self.body.posy += 1
                if height + self.body.posy > body_height:
                    self.body.resize(body_width, body_height + 1)
        elif key == caca.KEY_LEFT:
            x -= 1
            if x < 0:
                x = 0
                self.body.posx -= 1
                if self.body.posx < 0: self.body.posx = 0
        elif key == caca.KEY_RIGHT:
            x += 1
            if x > width:
                x = width
                self.body.posx += 1
                if width + self.body.posx > body_width:
                    self.body.resize(body_width + 1, body_height)

        #move cursor
        self.cursor.set_position(x, y)

    def statusbar_update(self):
        """ Callback method to display statusbar messages.
        """
        #right message
        mode_list = [
            'OPENED', 'IMPORT', 'DRAWTEXT', 'DRAWLINE',
            'DRAWBOX', 'DRAWTRIANGLE1', 'DRAWTRIANGLE2',
            'DRAWELLIPSE1', 'DRAWELLIPSE2', 'DRAWCIRCLE',
            'MOVETEXT', 'MOVELINE', 'MOVEBOX', 'MOVETRIANGLE',
            'MOVECIRCLE', 'MOVEELLIPSE',
        ]

        if self.mode() in mode_list:
            x, y = self.cursor.where()
            self.status.right = "%s x:%d y:%d" % (self.body.shortname(),
                                                  x+self.body.posx,
                                                  y+self.body.posy-1)
        else:
            self.status.right = None

        #left message
        if cacaeditor.INFO_MSG.has_key(self.mode()):
            self.status.left = cacaeditor.INFO_MSG[self.mode()]

    def refresh(self):
        """ Refresh interface display.
        """
        #reset libcaca canvas
        self.cv.clear()

        #draw edit buffer
        if self.body is not None:
            data = self.body.cv.export_area_to_memory(self.body.posx,
                                                      self.body.posy,
                                                      self.cv.get_width()-1,
                                                      self.cv.get_height()-2,
                                                      "caca")
            self.cv.import_area_from_memory(0, 1, data, "caca")

        #change colors
        self.colors.change()

        #draw object
        if self.drawobj is not None:
            self.show_object()

        #restore default colors
        self.colors.default()

        #draw menu
        if self.menu is not None:
            self.cv.blit(0, 0, self.menu(), NullCanvas())

        #draw submenu
        if self.mode() in ('READY', 'OPENED'):
            if self.menu.idx > 0:
                self.cv.blit(self.menu.getx(), 1,
                             self.menu.subs[self.menu.idx-1](), NullCanvas())

        #draw statusbar
        self.statusbar_update()
        if self.status is not None:
            self.cv.blit(0, self.cv.get_height()-1, self.status(), NullCanvas())

        #draw dialog box
        if self.dialog is not None:
            posx, posy = self.center(self.dialog.cv)
            self.cv.blit(posx, posy, self.dialog(), NullCanvas())

        #draw error dialog
        if self.errdialog is not None:
            posx, posy = self.center(self.errdialog.cv)
            self.cv.blit(posx, posy, self.errdialog(), NullCanvas())

        #refresh display
        self.dp.refresh()

    def show_object(self):
        """ Show the current drawing object onto document.
        """
        #draw only object with flag
        if self.drawobj.flag:
            #show DrawText object
            if self.mode() in ('DRAWTEXT'):
                x, y = self.drawobj.get_start_position()

                #draw each line from start position
                lines = self.drawobj().split('\n')
                for line in lines:
                    self.cv.put_str(x-self.body.posx, y-self.body.posy, line)
                    y += 1

            #show DrawText canvas
            elif self.mode() in ('MOVETEXT', 'TRANSFORM'):
                x, y = self.drawobj.get_start_position()

                #blit the DrawText canvas with transformations
                cv = self.drawobj.make_cv(self.colors.fg, self.colors.bg)
                self.cv.blit(x-self.body.posx, y-self.body.posy, cv)

            #show DrawLine object
            elif self.mode() in ('DRAWLINE', 'MOVELINE'):
                #set current drawing object
                obj = self.drawobj

                #compute new object position
                x1, x2 = obj.x1 - self.body.posx, obj.x2 - self.body.posx
                y1, y2 = obj.y1 - self.body.posy, obj.y2 - self.body.posy

                #draw the line
                if obj.thin:
                    self.cv.draw_thin_line(x1, y1, x2, y2)
                else:
                    self.cv.draw_line(x1, y1, x2, y2, obj.ch)

            #show DrawBox object
            elif self.mode() in ('DRAWBOX', 'MOVEBOX'):
                #set current drawing object
                obj = self.drawobj

                #compute new object width
                if obj.x1 >= obj.x2:
                    width = obj.x1 - obj.x2
                    x = obj.x2 - self.body.posx
                elif obj.x1 < obj.x2:
                    width = obj.x2 - obj.x1
                    x = obj.x1 - self.body.posx

                #compute new object height
                if obj.y1 >= obj.y2:
                    height = obj.y1 - obj.y2
                    y = obj.y2 - self.body.posy
                elif obj.y1 < obj.y2:
                    height = obj.y2 - obj.y1
                    y = obj.y1 - self.body.posy

                #draw the box
                if obj.thin:
                    self.cv.draw_thin_box(x, y, width+1, height+1)
                elif obj.fill:
                    self.cv.fill_box(x, y, width+1, height+1, obj.ch)
                else:
                    self.cv.draw_box(x, y, width+1, height+1, obj.ch)

            #show DrawTriangle object
            elif self.mode() in ('DRAWTRIANGLE2', 'MOVETRIANGLE'):
                #set current drawing object
                obj = self.drawobj

                #compute screen object position
                x1, y1 = obj.x1-self.body.posx, obj.y1-self.body.posy
                x2, y2 = obj.x2-self.body.posx, obj.y2-self.body.posy
                x3, y3 = obj.x3-self.body.posx, obj.y3-self.body.posy

                #draw the triangle
                if obj.thin:
                    self.cv.draw_thin_triangle(x1, y1, x2, y2, x3, y3)
                elif obj.fill:
                    self.cv.fill_triangle(x1, y1, x2, y2, x3, y3, obj.ch)
                else:
                    self.cv.draw_triangle(x1, y1, x2, y2, x3, y3, obj.ch)

            #show DrawCircle object
            elif self.mode() in ('DRAWCIRCLE', 'MOVECIRCLE'):
                obj = self.drawobj
                x, y = obj.x-self.body.posx, obj.y-self.body.posy
                self.cv.draw_circle(x, y, obj.radius, obj.ch)
            elif self.mode() in ('DRAWELLIPSE2', 'MOVEELLIPSE'):
                obj = self.drawobj
                x, y = obj.x-self.body.posx, obj.y-self.body.posy
                if obj.thin:
                    self.cv.draw_thin_ellipse(x, y, obj.ra, obj.rb)
                elif obj.fill:
                    self.cv.fill_ellipse(x, y, obj.ra, obj.rb, obj.ch)
                else:
                    self.cv.draw_ellipse(x, y, obj.ra, obj.rb, obj.ch)

            #show DrawImport object
            elif self.mode() == 'IMPORT':
                obj = self.drawobj
                self.cv.blit(obj.x, obj.y, obj.cv, NullCanvas())

    # Events
    #
    # Dispatch all bind events
    #

    def dispatch_keyboard_event(self, ev):
        """ Method to dispatch keyboard event.

            :ev: libcaca Event object
        """
        attributes = dir(self)
        attributes.sort()

        for item in attributes:
            if item.startswith('handle_'):
                ret = getattr(self, item)(ev)
                if ret:
                    if ret == cacaeditor.INTERRUPT:
                        debug(__name__, "%s called [interrupt]" % item)
                        break
                    elif ret == cacaeditor.CONTINUE:
                        debug(__name__, "%s called [continue]" % item)

    def dispatch_mouse_event(self, ev):
        """ Method to dispatch mouse event.

            :ev: libcaca Event object
        """
        pass

    def dispatch_resize_event(self, ev):
        """ Method to dispatch resize event.

            :ev: libcaca Event object
        """
        width, height = ev.get_resize_width(), ev.get_resize_height()
        debug(__name__, "%dx%d" % (width, height))
        self.menu.resize(width)
        self.status.resize(width)
        if self.body is not None:
            self.body.resize(width-1, height-2)

    # Callback methods
    #
    # Core utilities
    #

    def _open_document(self):
        """ Open a document into cacaeditor.
        """
        if self.dialog.answer:
            try:
                #store the filename
                filename = "".join(self.dialog.answer)

                #open document as a Buffer object
                self.body = Buffer(self.cv.get_width(),
                                   self.cv.get_height()-2, filename)

            except (cacaeditor.FileError, cacaeditor.AppError), err:
                self.mode.revert()
                self.dialog = None
                self.mode.set('ERRDIALOG')
                self.errdialog = MsgError(str(err))

            else:
                #revert cursor position
                x, y = self.cursor.old_position
                self.cursor.set_position(x, y)

                #change application mode
                self.mode.set('OPENED')

            #clean dialog
            self.dialog = None

        #cancel operation when no answer is available
        else:
            self.mode.revert()
            self.dialog = None

    def _save_document(self):
        """ Save the current document as filename (answer).
        """
        if self.dialog.answer:
            #save copy of document
            try:
                if self.body.save(filename="".join(self.dialog.answer)):
                    self.mode.revert()
            except cacaeditor.FileError, err:
                self.mode.revert()
                self.mode.set('ERRDIALOG')
                self.dialog = None
                self.errdialog = MsgError(str(err))

        x, y = self.cursor.old_position
        self.cursor.set_position(x, y)

        #clean dialog
        self.dialog = None

    def _choose_export(self):
        """ Store format for export.
        """
        if self.userdata[0]:
            #set application mode
            self.mode.set('OPENED', 'ANSWER')

            #ask for a filename for exporting data
            self.dialog = MsgDialog("Give a filename to export canvas:",
                                    callback=self._export_document,
                                    answer=True)

            #set cursor
            posx, posy = self.dialog.get_formxy(self.cv)
            self.cursor.set_position(posx, posy)
        else:
            #clean dialog
            self.dialog = None

    def _export_document(self):
        """ Export document in a file.
        """
        if self.dialog.answer and self.userdata[0]:
            #export data
            data = self.body.cv.export_to_memory(self.userdata[0])
            filename = "".join(self.dialog.answer)

            #clean dialog
            self.dialog = None
            self.userdata = []

            #write data to file
            try:
                with io.open(filename, 'wb+') as fp:
                    fp.write(data)
            except IOError, err:
                self.mode.set('OPENED', 'ERRDIALOG')
                self.errdialog = MsgError(str(err))
            else:
                #set application mode
                self.mode.set('OPENED')

    def _choose_import(self):
        """ Store format for import.
        """
        if self.userdata[0]:
            #set application mode
            self.mode.set('OPENED', 'ANSWER')

            #ask for filename to import
            self.dialog = MsgDialog("Give a filename to import:",
                                    callback=self._import_document,
                                    answer=True)

            #store cursor position
            x, y = self.cursor.where()
            self.cursor.old_position = x, y

            #set cursor position
            posx, posy = self.dialog.get_formxy(self.cv)
            self.cursor.set_position(posx, posy)

        else:
            #clean dialog
            self.dialog = None

    def _import_document(self):
        """ Import a file into document.
        """
        x, y = self.cursor.old_position

        if self.dialog.answer and self.userdata[0]:
            #set application mode
            self.mode.set('IMPORT')

            #import filename
            try:
                self.drawobj.load("".join(self.dialog.answer),
                                  fmt=self.userdata[0])
            except cacaeditor.FileError, err:
                #clean dialog
                self.dialog = None
                self.userdata = []

                self.mode.set('OPENED', 'ERRDIALOG')
                self.errdialog = MsgError(str(err))
                self.cursor.set_position(x, y)
                self.drawobj = None

        #skip empty answer
        else:
            self.mode.set('OPENED')
            self.drawobj = None

        self.cursor.set_position(x, y)

        #clean dialog
        self.dialog = None
        self.userdata = []

    def _input_char(self):
        """ Method to store character given.
        """
        #restore cursor origin
        x, y = self.cursor.old_position
        self.cursor.set_position(x, y)

        if self.dialog.answer:
            #change application mode
            self.mode.revert()
            self.drawobj.ch = self.dialog.answer[0]

        else:
            #reset application mode
            self.mode.set('OPENED')

            #clean drawing object
            self.drawobj = None

        #clean dialog
        self.dialog = None

