# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Graphical User Interface.
"""

#debug
#import pdb

#third party module
import caca

#cacaeditor modules
import cacaeditor
from cacaeditor.core import ModeApp
from cacaeditor.handles import ModeManager
from cacaeditor.widgets import Buffer, StatusBar, MenuBar
from cacaeditor.widgets import MsgListBox, MsgDialog, MsgError
from cacaeditor.helpers import DrawBox, DrawCircle, DrawEllipse
from cacaeditor.helpers import DrawLine, DrawText, DrawTriangle
from cacaeditor.helpers import DrawImport
from cacaeditor.helpers import debug, enable_mode

class App(ModeApp, ModeManager):
    """ Main class for cacaeditor.
    """
    def __init__(self, open_file=False, format_file=None, **kw):
        """ Overwrite application constructor.

            :open_file:   filename to open
            :format_file: format to use with open_file
        """
        #call __init__ from parent
        ModeApp.__init__(self, **kw)

        #open file given as argument
        if open_file:
            debug(__name__, "Open file given as argument: `%s'" % open_file)
            self.body = Buffer(self.cv.get_width(),
                               self.cv.get_height()-2)
            try:
                #load file with format
                if format_file:
                    debug(__name__, "Use format: `%s'" % format_file)
                    self.body.load(open_file, format_file)
                #try to autodetect format
                else:
                    self.body.load(open_file)
            except (cacaeditor.FileError, cacaeditor.AppError), err:
                self.mode.set('ERRDIALOG')
                self.errdialog = MsgError(str(err))
                self.cursor.display(False)
            else:
                self.cursor.set_position(0, 1)

                #change application mode
                self.mode.set('OPENED')

            #refresh display
            self.refresh()

    def build(self):
        """ Method to build user interface. This method is called when
            the object is instantiated.
        """
        #debug message
        debug(__name__, "Building interface")

        #set window title
        self.dp.set_title(self.name)

        #build menubar
        self.menu = MenuBar(self.cv.get_width(), cacaeditor.MENU_DATA)

        #build statusbar
        self.status = StatusBar(self.cv.get_width())

        #interface is now ready, change application mode
        self.mode.set('READY')
        debug(__name__, "Application is now ready")

        #refresh display
        self.refresh()

    #
    # Actions for 'File' menu
    #

    @enable_mode('READY')
    def action_new_document(self):
        """ Action method to create new document.
        """
        #init screen cursor
        self.cursor.set_position(0, 1)
        self.cursor.display(True)

        #create new document
        self.body = Buffer(self.cv.get_width()-1, self.cv.get_height()-2)

        #change application mode
        self.mode.set('OPENED')

    @enable_mode('READY')
    def action_open_document(self):
        """ Action method to open a document.
        """
        #init screen cursor
        self.cursor.set_position(0, 1)
        self.cursor.old_position = self.cursor.where()

        #change application mode
        self.mode.set('ANSWER')

        #ask for filename to import
        self.dialog = MsgDialog("Give the filename to open:",
                                callback=self._open_document,
                                answer=True)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_close_document(self):
        """ Action method to close a document.
        """
        #delete current document object
        self.body = None
        self.mode.set('READY')
        self.cursor.display(False)

    @enable_mode('OPENED')
    def action_save_document(self):
        """ Action method to save current document.
        """
        try:
            #save document if filename is set
            if self.body.filename is None:
                #otherwise call save as
                self.action_save_document_as()
            else:
                self.body.save()
        except cacaeditor.FileError, err:
            self.mode.set('ERRDIALOG')
            self.errdialog = MsgError(str(err))
            self.cursor.display(False)

    @enable_mode('OPENED')
    def action_save_document_as(self):
        """ Action method to save copy of document.
        """
        #store cursor position
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #ask for filename
        self.dialog = MsgDialog("Give the filename to save:",
                                callback=self._save_document,
                                answer=True)

        #change application mode
        self.mode.set('ANSWER')

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_export_document(self):
        """ Action method to export buffer.
        """
        self.mode.set('CHOOSEEXPORT')

        #ask for export format
        list_fmt = \
            ["%s: %s" % (fmt, desc) for fmt, desc in caca.get_export_list()]
        self.dialog = MsgListBox("Select the format for export:", list_fmt,
                                 callback=self._choose_export)
        self.cursor.display(False)

    @enable_mode('OPENED')
    def action_import_document(self):
        """ Action method to import data into buffer.
        """
        #init list format
        lfmt = []
        for fmt, desc in caca.get_import_list():
            if fmt != '""':
                lfmt.append("%s: %s" % (fmt, desc))

        #set application mode
        self.mode.set('CHOOSEIMPORT')

        #ask for import format
        self.dialog = MsgListBox("Select the format to import:", lfmt,
                              callback=self._choose_import)

        #init drawing object
        x, y = self.cursor.where()
        self.drawobj = DrawImport(x+self.body.posx, y+self.body.posy)
        self.cursor.display(True)

    @enable_mode('READY', 'OPENED')
    def action_quit_application(self):
        """ Action method to quit application.
        """
        #build window
        self.dialog = MsgDialog("Would you like to really quit ? [y/N]",
                                callback=self.destroy)
        self.mode.set('DIALOG')
        self.cursor.display(False)

    #
    # Actions for 'Draw' menu
    #

    @enable_mode('READY', 'OPENED', 'CHOOSECOLOR1', 'CHOOSECOLOR2')
    def action_change_colors(self):
        """ Action method to change foreground / background color.
        """
        #disable cursor
        self.cursor.display(False)

        if self.mode() in ('READY', 'OPENED'):
            #ask for color type
            self.dialog = MsgListBox("Which color do you want to change ?",
                                     ["foreground", "background"],
                                     callback=self.action_change_colors)

            #change application mode
            self.mode.set("CHOOSECOLOR1")
            self.menu.reset()

        elif self.mode() == "CHOOSECOLOR1":
            #ask for color
            color_names_k = [label for label in cacaeditor.COLOR_NAMES]
            self.dialog = MsgListBox("Select the color to apply ?",
                                     color_names_k,
                                     callback=self.action_change_colors)

            #change application mode
            self.mode.revert()
            self.mode.set("CHOOSECOLOR2")

        elif self.mode() == "CHOOSECOLOR2":
            #set new color
            if self.userdata[0] == "foreground":
                self.colors.fg = cacaeditor.COLOR_NAMES[self.userdata[1]]
            elif self.userdata[0] == "background":
                self.colors.bg = cacaeditor.COLOR_NAMES[self.userdata[1]]

            #change application mode
            self.mode.revert()

            #clean dialog
            self.dialog = None
            self.userdata = []

    @enable_mode('READY', 'OPENED')
    def action_reset_colors(self):
        """ Action method to reset colors to default.
        """
        self.colors.reset()

    @enable_mode('OPENED')
    def action_insert_text(self):
        """ Action method to insert text in buffer.
        """
        if self.drawobj is None:
            #set cursor origin
            x, y = self.cursor.where()
            self.cursor.old_position = (x, y)
            self.cursor.display(True)

            #build drawing object
            self.drawobj = DrawText(x + self.body.posx, y + self.body.posy)

            #change application mode
            self.mode.set('DRAWTEXT')

    @enable_mode('OPENED')
    def action_insert_line(self):
        """ Action method to draw a line.
        """
        #change application mode
        self.mode.set('DRAWLINE', 'ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to draw the line:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawLine(x+self.body.posx, y+self.body.posy)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_thin_line(self):
        """ Action method to draw a thin line.
        """
       #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)
        self.cursor.display(True)

        #create drawing object
        self.drawobj = DrawLine(x+self.body.posx, y+self.body.posy,
                                thin=True)

        #change application mode
        self.mode.set('DRAWLINE')

    @enable_mode('OPENED')
    def action_insert_box(self):
        """ Action method to draw a box.
        """
        #change application mode
        self.mode.set('DRAWBOX', 'ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to draw the box:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawBox(x+self.body.posx, y+self.body.posy)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_thin_box(self):
        """ Action method to draw a thin box.
        """
        #change application mode
        self.mode.set('DRAWBOX')

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)
        self.cursor.display(True)

        #create drawing object
        self.drawobj = DrawBox(x+self.body.posx, y+self.body.posy,
                               thin=True)

    @enable_mode('OPENED')
    def action_fill_box(self):
        """ Action method to fill a box.
        """
        #change application mode
        self.mode.set('DRAWBOX', 'ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to draw the box:",
                                callback=self._input_char,
                                answer=True)

        #store cursor position
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawBox(x+self.body.posx, y+self.body.posy,
                               fill=True)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_triangle(self):
        """ Action method to draw a triangle.
        """
        #change application mode
        self.mode.set('DRAWTRIANGLE1', 'ANSWER')

        #create dialog window
        self.dialog = MsgDialog("Give a character to draw the triangle:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawTriangle(x+self.body.posx, y+self.body.posy)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_thin_triangle(self):
        """ Action method to draw a thin triangle.
        """
        #change application mode
        self.mode.set('DRAWTRIANGLE1')

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)
        self.cursor.display(True)

        #create drawing object
        self.drawobj = DrawTriangle(x+self.body.posx, y+self.body.posy,
                                    thin=True)

    @enable_mode('OPENED')
    def action_fill_triangle(self):
        """ Action method to fill a triangle.
        """
        #change application mode
        self.mode.set('DRAWTRIANGLE1', 'ANSWER')

        #create dialog window
        self.dialog = MsgDialog("Give a character to draw the triangle:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawTriangle(x+self.body.posx, y+self.body.posy,
                                    fill=True)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_circle(self):
        """ Action method to draw a circle.
        """
        #change application mode
        self.mode.set('DRAWCIRCLE')
        self.mode.set('ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to draw the circle:",
                                callback=self._input_char,
                                answer=True)
        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawCircle(x+self.body.posx, y+self.body.posy)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_ellipse(self):
        """ Action method to draw an ellipse.
        """
        #change application mode
        self.mode.set('DRAWELLIPSE1', 'ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to draw an ellipse:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawEllipse(x+self.body.posx, y+self.body.posy)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    @enable_mode('OPENED')
    def action_insert_thin_ellipse(self):
        """ Action method to draw a thin ellipse.
        """
        #change application mode
        self.mode.set('DRAWELLIPSE1')

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)
        self.cursor.display(True)

        #create drawing object
        self.drawobj = DrawEllipse(x+self.body.posx, y+self.body.posy,
                                   thin=True)

    @enable_mode('OPENED')
    def action_fill_ellipse(self):
        """ Action method to fill an ellipse.
        """
        #change application mode
        self.mode.set('DRAWELLIPSE1', 'ANSWER')

        #create dialog
        self.dialog = MsgDialog("Give a character to fill ellipse:",
                                callback=self._input_char,
                                answer=True)

        #store cursor origin
        x, y = self.cursor.where()
        self.cursor.old_position = (x, y)

        #create drawing object
        self.drawobj = DrawEllipse(x+self.body.posx, y+self.body.posy,
                                   fill=True)

        #move cursor to answer form
        posx, posy = self.dialog.get_formxy(self.cv)
        self.cursor.set_position(posx, posy)
        self.cursor.display(True)

    #
    # Actions for 'Help' menu
    #

    @enable_mode('READY', 'OPENED', 'COMMAND')
    def action_list_format(self):
        """ Action method to display import and export supported
            formats.
        """
        #build import list
        msg = []
        msg.append("Import formats")
        msg.append("==============\n")
        for fmt, desc in caca.get_import_list():
            #skip autodetect
            if fmt != '""':
                msg.append("  * %10s  => %s" % (fmt, desc))
        msg.append("")

        #build export list
        msg.append("Export formats")
        msg.append("==============\n")
        for fmt, desc in caca.get_export_list():
            msg.append("  * %10s  => %s" % (fmt, desc))
        msg.append("")

        #build dialog window
        self.dialog = MsgDialog("\n".join(msg), width=70, wrap=False)
        self.mode.set('DIALOG')
        self.cursor.display(False)

    @enable_mode('READY', 'OPENED', 'COMMAND')
    def action_about(self):
        """ Action method to display About message box.
        """
        #build dialog window
        self.dialog = MsgDialog(cacaeditor.ABOUT_MSG, width=70, wrap=False)
        self.mode.set('DIALOG')
        self.cursor.display(False)

