# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Application modes manager.
"""

#debug
#import pdb

#third party module
import caca

#application modules
import cacaeditor
from cacaeditor.helpers import debug, enable_mode
from cacaeditor.widgets import MsgListBox

class ModeManager(object):
    """ Keybindings manager for application modes.
    """

    #
    # Global handles
    #

    @enable_mode('READY', 'OPENED')
    def handle_normal_mode(self, ev):
        """ Method to handle all keybindings when application was
            in READY or OPENED mode.

            :ev: libcaca Event object.
        """
        if self.mode() == 'OPENED':
            self.cursor.display(True)
        elif self.mode() == 'READY':
            self.cursor.display(False)

        #toggle menu
        if ev.get_key_ch() == caca.KEY_TAB:
            debug(__name__, "Select menu")
            self.menu.select()
            if self.menu.idx > 0:
                self.cursor.display(False)

        #clear menu
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            if self.menu.idx > 0:
                self.menu.reset()

        #activate menuitem
        elif ev.get_key_ch() == caca.KEY_RETURN:
            if self.menu.last_idx >= self.menu.idx > 0:
                subitem = self.menu.subs[self.menu.idx-1]
                if subitem.last_idx >= subitem.idx > 0:
                    tag = subitem.items[subitem.idx-1][0]
                    if hasattr(self, "action_%s" % tag):
                        getattr(self, "action_%s" % tag)()
                        debug(__name__, "action_%s triggered" % tag)
                    self.menu.reset()

        #navigate menu or document
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.menu.idx > 0:
                debug(__name__, "Select menu")
                if ev.get_key_ch() == caca.KEY_UP:
                    self.menu.subs[self.menu.idx-1].select(increment=-1)
                elif ev.get_key_ch() == caca.KEY_DOWN:
                    self.menu.subs[self.menu.idx-1].select()
                elif ev.get_key_ch() == caca.KEY_LEFT:
                    self.menu.select(increment=-1)
                elif ev.get_key_ch() == caca.KEY_RIGHT:
                    self.menu.select()
                if self.menu.idx > 0:
                    self.cursor.display(False)
            else:
                if self.mode() == 'OPENED':
                    self.move_cursor(ev.get_key_ch())

        #activate command mode
        elif ev.get_key_ch() == ord(':'):
            self.mode.set('COMMAND')
            self.status.left = "".join(self.prompt)
            x, y = self.cursor.where()
            self.cursor.old_position = (x, y)
            self.cursor.set_position(2, self.cv.get_height() - 1)
            self.cursor.display(True)

        #menu keybindings
        else:
            for menu in self.menu.data['_index']:
                for tag, desc, keybinding in self.menu.data[menu]:
                    if tag is not None and keybinding is not None:
                        k = keybinding.split('+')
                        if len(k) > 1:
                            if k[0] == "ctrl":
                                if ev.get_key_ch() == eval("caca.KEY_CTRL_%s" \
                                                % k[1].upper()):
                                    getattr(self, "action_%s" % tag)()
                                    debug(__name__, "action_%s triggered" \
                                          % tag)
                                    self.menu.reset()
                                    break
                        else:
                            if ev.get_key_ch() == ord(k[0]):
                                getattr(self, "action_%s" % tag)()
                                debug(__name__, "action_%s triggered" \
                                      % tag)
                                self.menu.reset()
                                break

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('ERRDIALOG')
    def handle_error_mode(self, ev):
        """ Handle all keybindings when application was
            in ERRDIALOG mode.
        """
        #clean error dialog
        debug(__name__, "Clean error dialog")
        self.errdialog = None
        self.mode.revert()
        if self.mode() != 'READY':
            self.cursor.display(True)

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('DIALOG')
    def handle_dialog_mode(self, ev):
        """ Method to handle all keybindings when application was
            in dialog mode ('DIALOG').

            :ev: libcaca Event object
        """
        #validate dialog
        if ev.get_key_ch() in cacaeditor.KEYS_YES:
            if self.dialog.callback is not None:
                self.dialog.callback()

        #clean dialog window
        debug(__name__, "Clean dialog")
        self.dialog = None
        self.mode.revert()
        if self.mode() != 'READY':
            self.cursor.display(True)

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('ANSWER')
    def handle_answer_mode(self, ev):
        """ Method to handle all keybindings when application was
            in answer mode ('ANSWER').

            :ev: libcaca Event object
        """

        #validate answer
        if ev.get_key_ch() == caca.KEY_RETURN:
            if self.dialog.callback is not None:
                debug(__name__, "Answer is `%s'" \
                      % "".join(self.dialog.answer))
                self.dialog.callback()

        #correct user answer
        elif ev.get_key_ch() == caca.KEY_BACKSPACE:
            if self.dialog.answer:
                utf8 = self.dialog.answer.pop()
                debug(__name__, "Remove `%s' from answer" % utf8)
                self.dialog.make()
                x, y = self.cursor.where()
                self.cursor.set_position(x-1, y)

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            debug(__name__, "Abort operation")
            self.cancel_action()
            self.dialog = None

        #input answer
        else:
            #append utf8 text
            utf8 = ev.get_key_utf8()
            #skip empty input
            if utf8 != '':
                debug(__name__, "Append `%s' to answer" % utf8)
                #clean utf8 buffer
                #hack for libcaca with ncurses driver in xterm
                if self.dp.get_driver() == 'ncurses':
                    utf8 = utf8.rstrip('\xff')
                self.dialog.answer.append(utf8)

            self.dialog.make()
            x, y = self.cursor.where()
            self.cursor.set_position(x+1, y)

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('CHOOSECOLOR1', 'CHOOSECOLOR2')
    def handle_choosecolor_mode(self, ev):
        """ Method to handle all keybindings when application was in
            mode choosecolor ('CHOOSECOLOR1', 'CHOOSECOLOR2').

            :ev: libcaca Event object
        """
        #validate user choice
        if ev.get_key_ch() == caca.KEY_RETURN:
            if self.dialog.idx > 0:
                if self.userdata:
                    debug(__name__, "Color %s is %s" \
                          % (self.userdata[0],
                             self.dialog.get_answer()))
                else:
                    debug(__name__, "Select %s" % self.dialog.get_answer())
                self.userdata.append(self.dialog.get_answer())
                self.dialog.callback()

        #navigate in list
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if ev.get_key_ch() == caca.KEY_UP:
                self.dialog.select(increment=-1)
            elif ev.get_key_ch() == caca.KEY_DOWN:
                self.dialog.select()

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.dialog = None
            self.userdata = []

        #end handle
        return cacaeditor.INTERRUPT

    #
    # Text handles
    #

    @enable_mode('DRAWTEXT')
    def handle_drawtext_mode(self, ev):
        """ Method to handle all keybindings when application was in
            DRAWTEXT mode.

            :ev: libcaca Event object
        """
        #validate user input
        if ev.get_key_ch() in cacaeditor.KEYS_VALIDATE:
            debug(__name__, "Validate text input: %s..." \
                  % str(self.drawobj.answer))
            self.mode.set('OPENED', 'MOVETEXT')

        #correct input with backspace
        elif ev.get_key_ch() == caca.KEY_BACKSPACE:
            if self.drawobj.answer:
                utf8 = self.drawobj.answer.pop()
                debug(__name__, "Remove `%s' from answer" % utf8)
                x, y = self.drawobj.get_end_position()
                self.cursor.set_position(x, y)

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #user input
        else:
            if ev.get_key_ch() == caca.KEY_RETURN:
                #append newline
                debug(__name__, "Append `\\n' to answer")
                self.drawobj.answer.append('\n')
                x, y = self.drawobj.get_end_position()
                self.cv.gotoxy(x, y)

            else:
                #append utf8 text
                utf8 = ev.get_key_utf8()

                #skip empty input
                if utf8 != '':
                    debug(__name__, "Append `%s' to answer" % utf8)
                    #clean utf8 buffer
                    #hack for libcaca with ncurses driver in xterm
                    if self.dp.get_driver() == 'ncurses':
                        utf8 = utf8.rstrip('\xff')
                    self.drawobj.answer.append(utf8)
                    self.move_cursor(caca.KEY_RIGHT)

        #interrupt keyboard dispatch
        return cacaeditor.INTERRUPT

    @enable_mode('MOVETEXT')
    def handle_movetext_mode(self, ev):
        """ Method to handle all keybindings when application was in
            MOVETEXT mode.

            :ev: libcaca Event object
        """
        #add/clear transorm operation
        if self.userdata:
            if self.userdata[0] == "normal":
                debug(__name__, "Reset transform list")
                self.drawobj.transform = []
            else:
                debug(__name__, "Append `%s' to transform list" \
                      % self.userdata[0])
                self.drawobj.transform.append(self.userdata[0])

            #clear dialog
            self.dialog = None
            self.userdata = []

            #interrupt handle
            return cacaeditor.INTERRUPT

        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            if self.drawobj.transform:
                debug(__name__, "Apply transform list: %s" \
                      % str(self.drawobj.transform))

            #draw input text to document
            x, y = self.drawobj.get_start_position()
            cv = self.drawobj.make_cv(color_fg=self.colors.fg,
                                      color_bg=self.colors.bg)
            self.body.blit(x, y, cv)

            #restore application mode
            self.mode.revert()

            #update cursor
            x, y = self.drawobj.get_end_position()
            self.cursor.set_position(x, y)

            #end user input
            self.drawobj = None

        #move text input
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())

        #ask for transformation list
        elif ev.get_key_ch() == caca.KEY_CTRL_T:
            self.dialog = MsgListBox("Select transformation to apply ?",
                                     cacaeditor.TRANSFORM_OPERATIONS,
                                     callback=self.handle_movetext_mode)
            self.cursor.display(False)
            self.mode.set('TRANSFORM')

        #cancel operation
        if ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #interrupt keyboard dispatch
        return cacaeditor.INTERRUPT

    @enable_mode('TRANSFORM')
    def handle_transform_mode(self, ev):
        """ Method to handle all keybindings when application was in
            TRANSFORM mode.

            :ev: libcaca Event object
        """

        #validate user choice
        if ev.get_key_ch() == caca.KEY_RETURN:
            if self.dialog.idx > 0:
                self.userdata.append(self.dialog.get_answer())
                last_mode = self.mode._last
                self.mode.set('OPENED', last_mode)
                self.dialog.callback(ev)

        #navigate in list
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if ev.get_key_ch() == caca.KEY_UP:
                self.dialog.select(increment=-1)
            elif ev.get_key_ch() == caca.KEY_DOWN:
                self.dialog.select()

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.dialog = None
            self.userdata = []
            self.mode.revert()

        #end handle
        return cacaeditor.INTERRUPT

    #
    # Drawing handles
    #

    @enable_mode('DRAWLINE')
    def handle_drawline_mode(self, ev):
        """ Method to handle all keybindings when application was
            in DRAWLINE mode.

            :ev: libcaca Event object
        """
        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            self.mode.set('OPENED', 'MOVELINE')

        #input end point
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            self.move_cursor(ev.get_key_ch())
            x, y = self.cursor.where()
            self.drawobj.set_last_point(x+self.body.posx, y+self.body.posy)
            debug(__name__, "Set last point: [x:%d y:%d]" \
                  % (x+self.body.posx, y+self.body.posy))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('MOVELINE')
    def handle_moveline_mode(self, ev):
        """ Method to handle all keybindings when application was
            in MOVELINE mode.

            :ev: libcaca Event object
        """
        #validate line
        obj = self.drawobj
        if ev.get_key_ch() == caca.KEY_RETURN:
            if obj.flag:
                #change colors
                self.body.colors.change(self.colors.fg, self.colors.bg)
                debug(__name__, "Draw line: [x:%d y:%d] [x:%d y:%d]" \
                      % (obj.x1, obj.y1, obj.x2, obj.y2))

                #draw line
                self.body.insert_line(obj.x1, obj.y1,
                                      obj.x2, obj.y2, obj.ch)

                #reset colors
                self.colors.default()

                #close drawing mode
                self.mode.revert()
                self.drawobj = None

        #move line
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())
                debug(__name__, "Move line: [x:%d y:%d] [x:%d y:%d]" \
                      % (obj.x1, obj.y1, obj.x2, obj.y2))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('DRAWBOX')
    def handle_drawbox_mode(self, ev):
        """ Method to handle all keybindings when application was
            in DRAWBOX mode.

            :ev: libcaca Event object
        """
        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            self.mode.set('OPENED', 'MOVEBOX')

        #edit input
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            self.move_cursor(ev.get_key_ch())
            x, y = self.cursor.where()
            self.drawobj.set_last_point(x+self.body.posx, y+self.body.posy)
            debug(__name__, "Set last point: [x:%d y:%d]" \
                  % (x+self.body.posx, y+self.body.posy))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('MOVEBOX')
    def handle_movebox_mode(self, ev):
        """ Method to handle all keybindings when application was
            in MOVEBOX mode.

            :ev: Libcaca Event object
        """
        if ev.get_key_ch() == caca.KEY_RETURN:
            obj = self.drawobj
            if obj.flag:
                #change colors
                self.body.colors.change(self.colors.fg, self.colors.bg)

                #draw the box
                x, y, width, height = self.drawobj.get_box_coords()
                self.body.insert_box(x, y, width, height,
                                           obj.ch, obj.fill)
                debug(__name__, "Draw box: [x:%d y:%d - %dx%d]" \
                      % (x, y, width, height))

                #restore colors
                self.body.colors.default()

                #close drawing mode
                self.drawobj = None
                self.mode.revert()

        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())
                x, y, width, height = self.drawobj.get_box_coords()
                debug(__name__, "Move box: [x:%d y:%d - %dx%d]" \
                      % (x, y, width, height))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('DRAWTRIANGLE1', 'DRAWTRIANGLE2')
    def handle_drawtriangle_mode(self, ev):
        """ Method to handle all keybindings when application was
            in mode drawtriangle ('DRAWTRIANGLE1', 'DRAWTRIANGLE2').

            :ev: libcaca Event object
        """
        #get cursor position
        x, y = self.cursor.where()

        if self.mode() == 'DRAWTRIANGLE1':
            #validate input
            if ev.get_key_ch() == caca.KEY_RETURN:
                self.mode.set('OPENED', 'DRAWTRIANGLE2')

            #update triangle
            elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
                self.move_cursor(ev.get_key_ch())
                self.drawobj.set_second_point(x+self.body.posx,
                                              y+self.body.posy)
                debug(__name__, "Set second point: [x:%d y:%d]" \
                      % (x+self.body.posx, y+self.body.posy))

        #second angle
        elif self.mode() == 'DRAWTRIANGLE2':
            #validate input
            if ev.get_key_ch() == caca.KEY_RETURN:
                self.mode.set('OPENED', 'MOVETRIANGLE')

            #update triangle
            elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
                self.move_cursor(ev.get_key_ch())
                self.drawobj.set_last_point(x+self.body.posx,y+self.body.posy)
                debug(__name__, "Set last point: [x:%d y:%d]" \
                      % (x+self.body.posx, y+self.body.posy))


        #cancel operation
        if ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('MOVETRIANGLE')
    def handle_movetriangle_mode(self, ev):
        """ Method to handle all keybindings when application was
            in MOVETRIANGLE mode.

            :ev: Libcaca Event object
        """
        if ev.get_key_ch() == caca.KEY_RETURN:
            obj = self.drawobj
            if obj.flag:
                #change colors
                self.body.colors.change(self.colors.fg, self.colors.bg)

                #draw object
                self.body.insert_triangle(obj.x1, obj.y1, obj.x2, obj.y2,
                                          obj.x3, obj.y3, obj.ch, obj.fill)
                debug(__name__, "Draw triangle: [x:%d y:%d] [x:%d y:%d]" \
                      "[x:%d y:%d]" % (obj.x1, obj.y1, obj.x2, obj.y2,
                                       obj.x3, obj.y3))

                #restore colors
                self.body.colors.default()

                #close drawing mode
                self.drawobj = None
                self.mode.revert()

        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())
                obj = self.drawobj
                debug(__name__, "Move triangle: [x:%d y:%d] [x:%d y:%d] " \
                      "[x:%d y:%d]" % (obj.x1, obj.y1, obj.x2, obj.y2,
                                       obj.x3, obj.y3))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('DRAWCIRCLE')
    def handle_drawcircle_mode(self, ev):
        """ Method to handle all keybindings when application was
            in mode drawcircle ('DRAWCIRCLE').

            :ev: libcaca Event object
        """
        if ev.get_key_ch() == caca.KEY_RETURN:
            self.mode.set('OPENED', 'MOVECIRCLE')

        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            self.move_cursor(ev.get_key_ch())
            x, y = self.cursor.where()
            self.drawobj.set_radius(x+self.body.posx, y+self.body.posy)
            debug(__name__, "Set radius: [x:%d y:%d]" \
                  % (x+self.body.posx, y+self.body.posy))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('MOVECIRCLE')
    def handle_movecircle_mode(self, ev):
        """ Method to handle all keybindings when application was
            in MOVECIRCLE mode.

            :ev: Libcaca Event object
        """
        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            obj = self.drawobj
            if obj.flag:
                #change colors
                self.body.colors.change(self.colors.fg, self.colors.bg)

                #draw object
                self.body.insert_circle(obj.x, obj.y, obj.radius, obj.ch)
                debug(__name__, "Draw circle: [x:%d y:%d radius:%.2f]" \
                      % (obj.x, obj.y, obj.radius))

                #restore colors
                self.body.colors.default()

                #close drawing mode
                self.mode.revert()
                self.drawobj = None

        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())
                obj = self.drawobj
                debug(__name__, "Move circle: [x:%d y:%d radius:%.2f]" \
                      % (obj.x, obj.y, obj.radius))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('DRAWELLIPSE1', 'DRAWELLIPSE2')
    def handle_drawellipse_mode(self, ev):
        """ Method to handle all keybindings when application was
            in mode drawellipse ('DRAWELLIPSE1', 'DRAWELLIPSE2').

            :ev: libcaca Event object
        """
        #get cursor position
        x, y = self.cursor.where()

        if self.mode() == 'DRAWELLIPSE1':
            #validate input
            if ev.get_key_ch() == caca.KEY_RETURN:
                #first radius
                self.mode.set('OPENED', 'DRAWELLIPSE2')

            #edit input
            elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
                self.move_cursor(ev.get_key_ch())
                self.drawobj.set_radius_a(x+self.body.posx, y+self.body.posy)
                debug(__name__, "Set radius a: [x:%d y:%d]" \
                      % (x+self.body.posx, y+self.body.posy))

        #second radius
        elif self.mode() == 'DRAWELLIPSE2':
            #validate input
            if ev.get_key_ch() == caca.KEY_RETURN:
                self.mode.set('OPENED', 'MOVEELLIPSE')

            elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
                self.move_cursor(ev.get_key_ch())
                self.drawobj.set_radius_b(x+self.body.posx, y+self.body.posy)
                debug(__name__, "Set radius b: [x:%d y:%d]" \
                      % (x+self.body.posx, y+self.body.posy))

        #cancel operation
        if ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('MOVEELLIPSE')
    def handle_moveellipse_mode(self, ev):
        """ Method to handle all keybindings when application was
            in MOVEELLIPSE mode.

            :ev: Libcaca Event object
        """
        if ev.get_key_ch() == caca.KEY_RETURN:
            obj = self.drawobj
            if obj.flag:
                #change colors
                self.body.colors.change(self.colors.fg, self.colors.bg)

                #draw object
                self.body.insert_ellipse(obj.x, obj.y, obj.ra, obj.rb,
                                         obj.ch, obj.fill)
                debug(__name__, "Draw ellipse: [x:%d y:%d radius_a:%d "
                      "radius_b:%d]" % (obj.x, obj.y, obj.ra, obj.rb))

                #restore colors
                self.body.colors.default()

                #close drawing mode
                self.drawobj = None
                self.mode.revert()

        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.drawobj.move(ev.get_key_ch()):
                self.move_cursor(ev.get_key_ch())
                obj = self.drawobj
                debug(__name__, "Draw ellipse: [x:%d y:%d radius_a:%d "
                      "radius_b:%d]" % (obj.x, obj.y, obj.ra, obj.rb))

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.cancel_action()
            self.drawobj = None

        #end handle
        return cacaeditor.INTERRUPT

    #
    # Import / Export handles
    #

    @enable_mode('IMPORT', 'CHOOSEIMPORT')
    def handle_import_mode(self, ev):
        """ Method to handle all keybindings when application was
            in mode import ('CHOOSEIMPORT', 'IMPORT').

            :ev: libcaca Event object
        """
        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            #choose format
            if self.mode() == 'CHOOSEIMPORT':
                if self.dialog.idx > 0:
                    self.userdata.append(self.dialog.get_answer().split(':')[0])
                    self.dialog.callback()

            #import data
            elif self.mode() == 'IMPORT':
                data = self.drawobj
                if data.flag:
                    self.body.blit(data.x, data.y-1, data.cv)
                    #close import mode
                    self.drawobj = None
                    self.mode.set('OPENED')

        #edit import position
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            #navigate into list
            if self.mode() == 'CHOOSEIMPORT':
                if ev.get_key_ch() == caca.KEY_UP:
                    self.dialog.select(increment=-1)
                elif ev.get_key_ch() == caca.KEY_DOWN:
                    self.dialog.select()

            #move data
            elif self.mode() == 'IMPORT':
                self.move_cursor(ev.get_key_ch())
                self.drawobj.x, self.drawobj.y = self.cursor.where()

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.mode.set('OPENED')
            self.cursor.set_position(self.drawobj.x, self.drawobj.y)
            self.drawobj = None
            self.dialog = None

        #end handle
        return cacaeditor.INTERRUPT

    @enable_mode('EXPORT', 'CHOOSEEXPORT')
    def handle_export_mode(self, ev):
        """ Method to handle all keybindings when application was
            in export mode ('CHOOSEEXPORT', 'EXPORT').

            :ev: Libcaca Event object
        """
        #validate input
        if ev.get_key_ch() == caca.KEY_RETURN:
            if self.mode() == 'CHOOSEEXPORT':
                if self.dialog.idx > 0:
                    self.userdata.append(self.dialog.get_answer().split(':')[0])
                    self.dialog.callback()

        #choose format
        elif ev.get_key_ch() in cacaeditor.KEYS_ARROW:
            if self.mode() == 'CHOOSEEXPORT':
                if ev.get_key_ch() == caca.KEY_UP:
                    self.dialog.select(increment=-1)
                elif ev.get_key_ch() == caca.KEY_DOWN:
                    self.dialog.select()

        #cancel operation
        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.mode.set('OPENED')
            if self.drawobj is not None:
                self.cursor.set_position(self.drawobj.x, self.drawobj.y)
            self.drawobj = None
            self.dialog = None

        #end handle
        return cacaeditor.INTERRUPT

    #
    # Command handles
    #

    @enable_mode('COMMAND')
    def handle_command_mode(self, ev):
        """ Method to handle all keybindings when application was
            in COMMAND mode.

            :ev: Libcaca Event project
        """
        if ev.get_key_ch() == caca.KEY_RETURN:
            cmd_list = [item.cmd for item in self.commands]
            cmd = "".join(self.prompt[2:])
            cmd_args = cmd.split(' ')[1:]
            cmd = cmd.split(' ')[0]
            if cmd in cmd_list:
                cmd_idx = cmd_list.index(cmd)
                getattr(self.commands[cmd_idx], "launch")(*cmd_args)
                debug(__name__, "%s COMMAND with args: %s" \
                      % (cmd.upper(), str(cmd_args)))
            else:
                debug(__name__, "UNKNOWN COMMAND %s" % cmd.upper())

                #reset application mode
                self.mode.revert()
                x, y = self.cursor.old_position
                self.cursor.set_position(x, y)

            #clear prompt
            self.prompt = ['>', ' ']

        elif ev.get_key_ch() == caca.KEY_BACKSPACE:
            if len(self.prompt) > 2:
                utf8 = self.prompt.pop()
                debug(__name__, "Remove `%s' from prompt" % utf8)
                self.cursor.set_position(len(self.prompt),
                                         self.cv.get_height()-1)

        elif ev.get_key_ch() in cacaeditor.KEYS_ABORT:
            self.mode.revert()
            self.prompt = ['>', ' ']
            x, y = self.cursor.old_position
            self.cursor.set_position(x, y)
            if self.mode() == 'READY':
                self.cursor.display(False)

        else:
            utf8 = ev.get_key_utf8()
            if utf8 != '':
                #clean utf8 buffer
                #hack for libcaca with ncurses driver in xterm
                if self.dp.get_driver() == 'ncurses':
                    utf8 = utf8.rstrip('\xff')
                debug(__name__, "Append `%s' to prompt" % utf8)
                self.prompt.append(utf8)
                self.cursor.set_position(len(self.prompt),
                                         self.cv.get_height()-1)

        self.status.left = "".join(self.prompt)

        return cacaeditor.INTERRUPT

