# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Utilities for ascii editor.
"""

#debug
#import pdb

#standard module
import math
import io
import sys
import traceback

#third party modules
import caca
from caca.canvas import Canvas, CanvasError

#application module
import cacaeditor

#
# Functions
#

def debug(module, message):
    """ Common function to print debug messages.

        :module: module __name__
        :message: the message to display
    """
    if cacaeditor.DEBUG:
        #skip unwanted modules
        if module.split('.')[-1] in cacaeditor.DEBUG_MODULES:
            #catch frame
            fid, frame = sys._current_frames().items()[-1]

            #extract parent info from stack
            filename, lineno, name, line = traceback.extract_stack(frame)[-2]

            #display debug message
            print("DEBUG:%s:%s: %s" % (module, name, message))

def increment_xy(key):
    """ Return increment for coords.

        :key: the arrow key (left, right, up, down).
    """
    inc_x = inc_y = 0

    if key == caca.KEY_LEFT:
        inc_x = -1
    elif key == caca.KEY_RIGHT:
        inc_x = 1
    elif key == caca.KEY_UP:
        inc_y = -1
    elif key == caca.KEY_DOWN:
        inc_y = 1

    return (inc_x, inc_y)

#
# Decorators
#

def enable_mode(*enabled):
    """ Application mode filter decorator for gui methods.

        :enabled: valid mode for application
    """
    #method level wrapper
    def wrapper(func):
        #class level wrapper
        def wrapped(self, *args, **kw):
            #filter current mode in args list
            if self.mode() in enabled:
                #execute method
                return func(self, *args, **kw)
            else:
                #skip method execution
                return False
        return wrapped
    return wrapper

#
# Application utilites
#

class ColorObject(object):
    """ Object to store foreground and background color value of application.

         _fg, _bg => default colors
          fg,  bg => current colors
    """
    __slots__ = ('_fg', '_bg', 'fg', 'bg', '__dict__')

    def __init__(self, ref_cv, value_fg, value_bg):
        """ ColorObject constructor.

            :ref_cv: reference canvas for application
            :value_fg: default color value for foreground
            :value_bg: default color value for background
        """
        self._fg, self._bg = value_fg, value_bg
        self.fg, self.bg = value_fg, value_bg
        self._set_color_ansi = ref_cv.set_color_ansi

    def change(self, fg=None, bg=None):
        """ Change the current colors for application.

            :fg: new color value for foreground
            :bg: new color value for background
        """
        if fg is not None:
            self.fg = fg
        if bg is not None:
            self.bg = bg

        self._set_color_ansi(self.fg, self.bg)

    def default(self):
        """ Set default colors for application.
        """
        self._set_color_ansi(self._fg, self._bg)

    def reset(self):
        """ Reset foreground and background colors to default.
        """
        if self.fg != self._fg:
            self.fg = self._fg
        if self.bg != self._bg:
            self.bg = self._bg

        self._set_color_ansi(self.fg, self.bg)

class CursorObject(object):
    """ Object to store cursor position, x and y value.

        * x            => X coordinate for cursor.
        * y            => Y coordinate for cursor.
        * enable       => Boolean value that indicate cursor state.
        * old_position => Cursor coords as tuple
    """
    __slots__ = ('x', 'y', 'old_position', 'enable', '__dict__')

    def __init__(self, ref_cv, ref_dp, value_x, value_y):
        """ CursorObject constructor.

            :ref_cv:  the reference canvas for cursor
            :ref_dp:  the reference display for cursor
            :value_x: default value for X
            :value_y: default value for Y
        """
        #set current position
        self.x = int(value_x)
        self.y = int(value_y)

        #init object attributes
        self.old_position = (self.x, self.y)
        self.enable = False

        #methods inherited from canvas
        self._getx = ref_cv.wherex
        self._gety = ref_cv.wherey
        self._gotoxy = ref_cv.gotoxy

        #methods inherited from display
        self._display = ref_dp.set_cursor

    def display(self, value):
        """ Enable / disable cursor

            :value: boolean to indicate cursor state
        """
        if value and not self.enable:
            debug(__name__, "Enable cursor")
            self.enable = True
            self._display(self.enable)
        elif not value and self.enable:
            debug(__name__, "Disable cursor")
            self.enable = False
            self._display(self.enable)

    def set_position(self, value_x, value_y):
        """ Set the cursor position.

            :value_x: X coordinate for cursor position
            :value_y: Y coordinate for cursor position
        """
        self.x = int(value_x)
        self.y = int(value_y)

        self._gotoxy(self.x, self.y)

    def where(self):
        """ Return current cursor position as tuple.
        """
        return (self._getx(), self._gety())

class ModeObject(object):
    """ Object to store current application mode, and last mode set.
    """
    __slots__ = ('mode', '_last', '__dict__')

    def __init__(self):
        """ ModeObject constructor.
        """
        self._last = None
        self.mode = None

    def __call__(self):
        """ Return current application mode.
        """
        return self.mode

    def set(self, *args):
        """ Change current application mode.

            :args: mode to set for application
        """
        for arg in args:
            self._last = self.mode
            self.mode = arg

    def revert(self):
        """ Restore last application mode.
        """
        self.mode = self._last
        self._last = None

class CommandObject(object):
    """ Model class for command object.
    """
    cmd = None

    def __init__(self, app):
        """
        """
        self.app = app

    def launch(self, *args, **kw):
        """ Overwrite this method.
        """
        pass

    def help(self):
        """ Return docstring as help text.
        """
        doc = []
        for line in self.__doc__.split('\n'):
            doc.append(line.lstrip(' '))

        return "\n".join(doc)

    def is_on_canvas(self, x, y):
        """ Return True if coordinates are inside buffer canvas,
            otherwise return False.

            :X: coordinate to test
            :Y: coordinate to test
        """
        if x >= 0 and x < self.app.cv.get_width() - 1:
            if y >= 0 and y < self.app.cv.get_height() - 2:
                return True

        return False

#
# Drawing utilities
#

class DrawText(object):
    """ Container object to store the text to draw to screen.

        :x: the X coordinate for the start position
        :y: the Y coordinate for the start position
    """
    def __init__(self, x, y):
        """ DrawText constructor.
        """
        self.answer = []
        self.x = x
        self.y = y
        self.flag = True
        self.transform = []

    def __call__(self):
        """ Return answer as string.
        """
        return "".join(self.answer)

    def get_size(self):
        """ Return max with, and x y offset for cursor position.
        """
        yoffset = 1
        max_x = xoffset = 0
        for item in self.answer:
            xoffset += 1
            if xoffset > max_x:
                max_x = xoffset
            if item == '\n':
                yoffset += 1
                xoffset = 0

        return (max_x, xoffset, yoffset)

    def get_start_position(self):
        """ Return the cursor position at the start of the text.
        """
        return (self.x, self.y)

    def get_end_position(self):
        """ Return the cursor position at end of the text.
        """
        max_x, xoffset, yoffset = self.get_size()

        return (self.x + xoffset, self.y + yoffset - 1)

    def make_cv(self, color_fg=None, color_bg=None):
        """ Return canvas with text.

            :color_fg: the canvas foreground color
            :color_bg: the canvas background color
        """
        max_x, xoffset, yoffset = self.get_size()
        try:
            cv = Canvas(max_x, yoffset)
        except CanvasError, err:
            raise cacaeditor.AppError(err)

        if color_fg is not None or color_bg is not None:
            cv.set_color_ansi(color_fg, color_bg)

        y = 0
        for line in self.__call__().split('\n'):
            if line:
                cv.put_str(0, y, line)
                y += 1

        for t in self.transform:
            getattr(cv, t)()

        return cv

    def move(self, key):
        """ Move the text input along the canvas.

            :key: the arrow key
        """
        #set default return to True
        ret = True

        #get the increment position coords
        inc_x, inc_y = increment_xy(key)

        #increment x
        if inc_x != 0:
            self.x += inc_x
            if self.x < 0:
                self.x -= inc_x
                ret = False

        #increment y
        if inc_y != 0:
            self.y += inc_y
            if self.y < 1:
                self.y -= inc_y
                ret = False

        return ret

class DrawLine(object):
    """ Container object to store line coordinates.

        :x: the X coordinate of the first point
        :y: the Y coordinate of the first point
        :thin: use draw_thin_line if True
    """
    def __init__(self, x, y, thin=False, fill=False):
        """ DrawLine constructor.
        """
        #init line attributes
        self.x2 = None
        self.y2 = None
        self.ch = None
        self.flag = False

        #set first point
        self.x1 = x
        self.y1 = y

        #set line style
        self.thin = thin
        self.fill = fill

    def set_last_point(self, x, y):
        """ Set last point to end the line.

            :x: the X coordinate for the point
            :y: the Y coordinate for the point
        """
        self.x2 = x
        self.y2 = y

        #object is ready to draw
        self.flag = True

    def move(self, key):
        """ Move the line.

            :key: the arrow key
        """
        #set default return to True
        ret = True

        #get the increment position coords
        inc_x, inc_y = increment_xy(key)

        #increment x1, x2
        if inc_x != 0:
            self.x1 += inc_x
            self.x2 += inc_x
            if self.x1 < 0:
                self.x1 = 0
                self.x2 -= inc_x
                ret = False
            elif self.x2 < 0:
                self.x2 = 0
                self.x1 -= inc_x
                ret = False

        #increment y1, y2
        if inc_y != 0:
            self.y1 += inc_y
            self.y2 += inc_y
            if self.y1 < 1:
                self.y1 = 1
                self.y2 -= inc_y
                ret = False
            elif self.y2 < 1:
                self.y2 = 1
                self.y1 -= inc_y
                ret = False

        return ret

class DrawBox(DrawLine):
    """ Container object to store box coordinates.

        :x: the X coordinate for the upper left corner
        :y: the Y coordinate for the upper left corner
        :thin: use draw_thin_box if True
        :fill: use fill_box if True
    """
    def get_box_coords(self):
        """
        """
        if self.flag:
            if self.x1 > self.x2:
                width = self.x1 - self.x2
                x = self.x2
            else:
                width = self.x2 - self.x1
                x = self.x1
            if self.y1 > self.y2:
                height = self.y1 - self.y2
                y = self.y2
            else:
                height = self.y2 - self.y1
                y = self.y1

            return (x, y, width, height)
        else:
            return None

class DrawTriangle(object):
    """ Container object to store triangle coordinates.

        :x: X coordinate of the first point
        :y: Y coordinate of the first point
        :thin: use draw_thin_triangle if True
        :fill: use fill_triangle if True
    """
    def __init__(self, x, y, thin=False, fill=False):
        """ DrawTriangle constructor.
        """
        #init attributes
        self.x2 = None
        self.y2 = None
        self.x3 = None
        self.y3 = None
        self.ch = None
        self.flag = False

        #set first point
        self.x1 = x
        self.y1 = y

        #set line style
        self.thin = thin
        self.fill = fill

    def set_second_point(self, x, y):
        """ Set the second point of the triangle.

            :x: X coordinate of the second point
            :y: Y coordinate of the second point
        """
        self.x2 = x
        self.y2 = y

    def set_last_point(self, x, y):
        """ Set the third point of the triangle.

            :x: X coordinate of the third point
            :y: Y coordinate of the third point
        """
        self.x3 = x
        self.y3 = y

        #object is ready to draw
        self.flag = True

    def move(self, key):
        """ Move the triangle.
        """
        #set default return to True
        ret = True

        #get the increment position coords
        inc_x, inc_y = increment_xy(key)

        #increment x1, x2
        if inc_x != 0:
            self.x1 += inc_x
            self.x2 += inc_x
            self.x3 += inc_x
            if self.x1 < 0:
                self.x1 = 0
                self.x2 -= inc_x
                self.x3 -= inc_x
                ret = False
            elif self.x2 < 0:
                self.x2 = 0
                self.x1 -= inc_x
                self.x3 -= inc_x
                ret = False
            elif self.x3 < 0:
                self.x3 = 0
                self.x1 -= inc_x
                self.x2 -= inc_x
                ret = False

        #increment y1, y2
        if inc_y != 0:
            self.y1 += inc_y
            self.y2 += inc_y
            self.y3 += inc_y
            if self.y1 < 1:
                self.y1 = 1
                self.y2 -= inc_y
                self.y3 -= inc_y
                ret = False
            elif self.y2 < 1:
                self.y2 = 1
                self.y1 -= inc_y
                self.y3 -= inc_y
                ret = False
            elif self.y3 < 1:
                self.y3 = 1
                self.y1 -= inc_y
                self.y2 -= inc_y
                ret = False

        return ret

class DrawCircle(object):
    """ Container object to store circle coordinates.

        :x: X coordinate of the center
        :y: Y coordinate of the center
    """
    def __init__(self, x, y):
        """ DrawCircle constructor.
        """
        #init attributes
        self.ch = None
        self.radius = None
        self.flag = False

        #set center of circle
        self.x = x
        self.y = y

    def set_radius(self, x, y):
        """ Set the circle radius.

            :x: X coordinate for the point to define radius
            :y: Y coordinate for the point to define radius
        """
        ax, ay = x - self.x, y - self.y
        self.radius = int(round(math.sqrt((ax * ax) + (ay * ay))))

        #object is ready to draw
        self.flag = True

    def move(self, key):
        """ Move the circle.
        """
        #set default return to True
        ret = True

        #get the increment position coords
        inc_x, inc_y = increment_xy(key)

        #increment x, y
        if inc_x != 0:
            self.x += inc_x
            if self.x < 0:
                self.x = 0
                ret = False
        if inc_y != 0:
            self.y += inc_y
            if self.y < 1:
                self.y = 1
                ret = False

        return ret

class DrawEllipse(object):
    """ Container object to store ellipse coordinates.

        :x:     X coordinate of the center
        :y:     Y coordinate of the center
        :thin:  if True use draw_thin_ellispe
    """
    def __init__(self, x, y, thin=False, fill=False):
        """ DrawEllipse constructor.
        """
        #init attributes
        self.ra = None
        self.rb = None
        self.ch = None
        self.flag = False
        self.fill = False

        #set ellipse center
        self.x = x
        self.y = y

        #set ellipse style
        self.thin = thin
        self.fill = fill

    def set_radius_a(self, x, y):
        """ Set the primary radius.

            :x: X coordinate for the point to define radius
            :y: Y coordinate for the point to define radius
        """
        ax, ay = x - self.x, y - self.y
        self.ra = int(round(math.sqrt((ax * ax) + (ay * ay))))

    def set_radius_b(self, x, y):
        """ Set the second radius.

            :x: X coordinate for the point to define radius
            :y: Y coordinate for the point to define radius
        """
        bx, by = x - self.x, y - self.y
        self.rb = int(round(math.sqrt((bx * bx) + (by * by))))

        #object is ready to draw
        self.flag = True

    def move(self, key):
        """ Move ellipse.
        """
        #set default return to True
        ret = True

        #get the increment position coords
        inc_x, inc_y = increment_xy(key)

        #increment x, y
        if inc_x != 0:
            self.x += inc_x
            if self.x < 0:
                self.x = 0
                ret = False
        if inc_y != 0:
            self.y += inc_y
            if self.y < 1:
                self.y = 1
                ret = False

        return ret

class DrawImport(object):
    """ Container object to store imported datas canvas.
    """
    def __init__(self, x, y):
        """ DataImport constructor.
        """
        self.x = x
        self.y = y
        self.flag = False

        try:
            self.cv = Canvas()
        except CanvasError, e:
            raise cacaeditor.AppError(e)

    def load(self, filename, fmt=None):
        """ Import data into canvas.

            :filename: the filename to load
        """
        try:
            with io.open(filename, 'rb') as fp:
                try:
                    if fmt is None:
                        self.cv.import_from_memory(fp.read(), "")
                    else:
                        self.cv.import_from_memory(fp.read(), fmt)
                except CanvasError, err:
                    raise cacaeditor.AppError(str(err))
                else:
                    self.flag = True
        except IOError, err:
            raise cacaeditor.FileError(str(err))

